/**
 * ================================================
 *  Gnome Shell Extension - Hide Titlebar - README 
 * ================================================
 * 
 * The gnome shell extension "Hide Titlebar" is a fork of the extension
 * "Maximus", which allows to hide the titlebars of maximized windows. Unlike
 * the original extension it is possible, that only the titlebars of windows,
 * which are located on the screen, that contains the top panel of the GNOME
 * shell, are hidden by the extension due to better multi-monitor support.
 * 
 * The original extension "Maximus" is available on the official gnome shell
 * extensions-homepage or on Bitbucket via the following links. Please
 * appreciate the work of the orignal author Amy Chan!
 * 
 * GNOME Shell Extensions: https://extensions.gnome.org/extension/354/maximus
 * Bitbucket:			   https://bitbucket.org/mathematicalcoffee/maximus-gnome-shell-extension
 * 
 * This fork of the extension above is available on Sourceforge via the
 * following URL:
 * 
 * https://sourceforge.net/projects/hide-titlebar
 * 
 * The aim of developing this extension was to provide a possibility to use the
 * top panel of the GNOME shell as the title bar of a window, which is maximized
 * on the appropriate screen. Therefore the title bar of the window should be
 * hidden when it is maximized and the window's buttons and title should be
 * accessible via the top panel. With the gnome shell extensions, that already
 * existed, this was only possible for setups, that use one screen, but not for
 * multi-monitor setups, because the extensions offered no ability to handle the
 * windows on screens that contain no top panel differently. Therefore a fork of
 * the original extension was created to add this additional feature. Via the
 * preferences dialog of the extension is it possible to turn off/on, whether
 * only windows on the screen, that contains the top panel, should be considered
 * by the extension. When turning this option off the behaviour of the extension
 * will be equal to the original one. Furthermore it is possible to choose the
 * screen, which contains the top panel to support setups, that are configured
 * to show the panel on an other screen but the primary one.
 * 
 * Besides this extension, also other extensions have been created as forks of
 * existing ones to reach the aim explained above. If you want to configure your
 * GNOME shell to behave like this, it is recommended to use the following
 * extensions in addition to the one this readme belongs to:
 * 
 *  "Title to Panel"
 * ------------------
 * 
 * This extension is a fork of the extension "Status Title Bar" and allows to
 * display the title of the currently selected window in the top panel of the
 * GNOME shell. Such like when using this extension it is possible that only
 * windows, that are located on the screen, which contains the top panel, are
 * considered by the application. Furthermore there is also an option, that
 * allows to only show the titles of maximized windows in the panel. These
 * options in the extension's preferences dialog should be enabled. The
 * extension is available on Sourceforge via the following link:
 * 
 * https://sourceforge.net/projects/title-to-panel
 * 
 *  "Buttons to Panel"
 * --------------------
 * 
 * This extension is a fork of the extension "Window Buttons" and allows to show
 * the window buttons of the currently focused window in the top panel of the
 * GNOME shell. The extension also allows to set, whether windows, that are
 * located on an other screen than the one, which contains the top panel, should
 * be taken into consideration. The appropriate option in the extension's
 * preferences dialog should be enabled. Also the option "any window maximized"
 * should be selected to define, when the buttons should be displayed. The
 * extension is available on Sourceforge via the following link:
 * 
 * https://sourceforge.net/projects/buttons-to-panel/
 * 
 * ==============
 *  Installation
 * ==============
 * 
 * For using this extension it is necessary to copy the folder
 * "hide-titlebar@mrapp.sourceforge.com", which is contained by the package, the
 * extension has been shiped with, to the path
 * "~/.local/share/gnome-shell/extensions". For installing the extension this
 * way, the shell script INSTALL.sh can be used. This script automatically
 * copies all relevant files to the appropriate folder and replaces an already
 * installed version of the extension when existing. When using the bash script,
 * the installation of this extension is possible by running the following
 * commands in a terminal (the name of the zip-file maybe has to be adjusted to
 * the actual version of the release):
 * 
 * $ unzip gnome-shell-extension-hide-titlebar-v.1.0.zip 
 * $ cd gnome-shell-extension-hide-titlebar 
 * $./INSTALL.sh
 * 
 * To enable the extension or to access its preferences afterwards, a program
 * such like "Gnome Tweak Tool" can be used.
 * 
 * 
 * ==============
 *  Known Issues
 * ==============
 *
 * When using the option "Only consider windows on the screen, that contains the 
 * panel" and a window is moved to an other screen programmatically, e.g. by 
 * using a keyboard shortcut, the title bar maybe will not be shown/hidden 
 * properly.
 *
 * ======== 
 *  Author 
 * ========
 * 
 * Michael Rapp (E-Mail: michael.rapp90@googlemail.com)
 * 
 * This software has been published as open source according to the terms and
 * conditions of the GNU General Public License Version 2.
 */
const GLib = imports.gi.GLib;
const Lang = imports.lang;
const Mainloop = imports.mainloop;
const Meta = imports.gi.Meta;
const Shell = imports.gi.Shell;
const Util = imports.misc.util;

const Main = imports.ui.main;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;
const Prefs = Me.imports.prefs;

// convenience
Meta.MaximizeFlags.BOTH = (Meta.MaximizeFlags.VERTICAL | Meta.MaximizeFlags.HORIZONTAL);

let maxID = null, minID = null, settingsChangedID = null, changeWorkspaceID = null,
		grabID = null;
let workspaces = [];
let oldFullscreenPref = null;
let settings = null;
let onetime = 0;
let APP_LIST, IS_BLACKLIST, USE_SET_HIDE_TITLEBAR;

function LOG(message) {
	// log(message);
}

/**
 * Guesses the X ID of a window.
 * 
 * It is often in the window's title, being `"0x%x %10s".format(XID,
 * window.title)`. (See `mutter/src/core/window-props.c`).
 * 
 * If we couldn't find it there, we use `win`'s actor,
 * `win.get_compositor_private()`. The actor's `x-window` property is the X ID
 * of the window *actor*'s frame (as opposed to the window itself).
 * 
 * However, the child window of the window actor is the window itself, so by
 * using `xwininfo -children -id [actor's XID]` we can attempt to deduce the
 * window's X ID.
 * 
 * It is not always foolproof, but works good enough for now.
 * 
 * @param {Meta.Window}
 *            win - the window to guess the XID of. You wil get better success
 *            if the window's actor (`win.get_compositor_private()`) exists.
 */
function guessWindowXID(win) {
	let	id = null;
	/*
	 * if window title has non-utf8 characters, get_description() complains
	 * "Failed to convert UTF-8 string to JS string: Invalid byte sequence in
	 * conversion input", event though get_title() works.
	 */
	try {
		id = win.get_description().match(/0x[0-9a-f]+/);
		if (id) {
			id = id[0];
			return id;
		}
	} catch (err) {
	
	}

	// use xwininfo, take first child.
	let	act = win.get_compositor_private();
	if (act) {
		id = GLib.spawn_command_line_sync('xwininfo -children -id 0x%x'
				.format(act['x-window']));
		if (id[0]) {
			let str = id[1].toString();

			/*
			 * The X ID of the window is the one preceding the target window's
			 * title. This is to handle cases where the window has no frame and
			 * so act['x-window'] is actually the X ID we want, not the child.
			 */
			let regexp = new RegExp('(0x[0-9a-f]+) +"%s"'.format(win.title));
			id = str.match(regexp);
			if (id) {
				return id[1];
			}

			/* Otherwise, just grab the child and hope for the best */
			id = str.split(/child(?:ren)?:/)[1].match(/0x[0-9a-f]+/);
			if (id) {
				return id[0];
			}
		}
	}
	// debugging for when people find bugs..
	log("[hide-titlebar]: Could not find XID for window with title %s"
			.format(win.title));
	return null;
}

/**
 * Undecorates a window.
 * 
 * If I use set_decorations(0) from within the GNOME shell extension (i.e. from
 * within the compositor process), the window dies. If I use the same code but
 * use `gjs` to run it, the window undecorates properly.
 * 
 * Hence I have to make some sort of external call to do the undecoration. I
 * could use 'gjs' on a javascript file (and I'm pretty sure this is installed
 * with GNOME-shell too), but I decided to use a system call to xprop and set
 * the window's `_MOTIF_WM_HINTS` property to ask for undecoration.
 * 
 * We can use xprop using the window's title to identify the window, but prefer
 * to use the window's X ID (in case the title changes, or there are multiple
 * windows with the same title).
 * 
 * The Meta.Window object does *not* have a way to access a window's XID.
 * However, the window's description seems to have it. Alternatively, a window's
 * actor's 'x-window' property returns the XID of the window *frame*, and so if
 * we parse `xwininfo -children -id [frame_id]` we can extract the child XID
 * being the one we want.
 * 
 * See here for xprop usage for undecoration:
 * http://xrunhprof.wordpress.com/2009/04/13/removing-decorations-in-metacity/
 * 
 * @param {Meta.Window}
 *            win - window to undecorate.
 */
function undecorate(win) {
	/* Undecorate with xprop */
	let	id = guessWindowXID(win), cmd = [ 'xprop', '-id', id, '-f',
			'_MOTIF_WM_HINTS', '32c', '-set', '_MOTIF_WM_HINTS',
			'0x2, 0x0, 0x0, 0x0, 0x0' ];
	/*
	 * _MOTIF_WM_HINTS: see MwmUtil.h from OpenMotif source (cvs.openmotif.org),
	 * or rudimentary documentation here:
	 * http://odl.sysworks.biz/disk$cddoc04sep11/decw$book/d3b0aa63.p264.decw$book
	 * 
	 * Struct { flags, functions, decorations, input_mode, status }. Flags: what
	 * the hints are for. (functions, decorations, input mode and/or status).
	 * Functions: minimize, maximize, close, ... Decorations: title, border,
	 * all, none, ... Input Mode: modeless, application modal, system model, ..
	 * Status: tearoff window.
	 */

	// fallback: if couldn't get id for some reason, use the window's name
	if (!id) {
		cmd[1] = '-name';
		cmd[2] = win.get_title();
	}
	LOG(cmd.join(' '));
	Util.spawn(cmd);
	// #25: when undecorating a Qt app (texmaker, keepassx) somehow focus is
	// lost.
	// However, is there a use case where this would happen legitimately?
	// For some reaons the Qt apps seem to take a while to be refocused.
	Meta.later_add(Meta.LaterType.IDLE, function() {
		if (win.focus) {
			win.focus(global.get_current_time());
		} else {
			win.activate(global.get_current_time());
		}
	});
}

/**
 * Decorates a window by setting its `_MOTIF_WM_HINTS` property to ask for
 * decoration.
 * 
 * @param {Meta.Window}
 *            win - window to undecorate.
 */
function decorate(win) {
	/* Decorate with xprop: 1 == DECOR_ALL */
	let	id = guessWindowXID(win), cmd = [ 'xprop', '-id', id, '-f',
			'_MOTIF_WM_HINTS', '32c', '-set', '_MOTIF_WM_HINTS',
			'0x2, 0x0, 0x1, 0x0, 0x0' ];
	// fallback: if couldn't get id for some reason, use the window's name
	if (!id) {
		cmd[1] = '-name';
		cmd[2] = win.get_title();
	}
	LOG(cmd.join(' '));
	Util.spawn(cmd);
	// #25: when undecorating a Qt app (texmaker, keepassx) somehow focus is
	// lost.
	// However, is there a use case where this would happen legitimately?
	// For some reaons the Qt apps seem to take a while to be refocused.
	Meta.later_add(Meta.LaterType.IDLE, function() {
		if (win.focus) {
			win.focus(global.get_current_time());
		} else {
			win.activate(global.get_current_time());
		}
	});
}

/**
 * Fetches the properties of the screens, which are configured for the shell.
 */
function loadScreenData() {
    this._primary = global.screen.get_primary_monitor();
    let numMonitors = global.screen.get_n_monitors();
    this._screens = [];
  
    for (let i=0; i<numMonitors; i++) {
    	let geom = global.screen.get_monitor_geometry(i);
    	let primary = (i == this._primary);

    	this._screens[i] =  {
    		primary: primary,
            y: geom.y,
            x : geom.x,
	        geomX: geom.x,
            geomY: geom.y,
            totalWidth: geom.width,
            totalHeight: geom.height,
            east: [],
            west: [],
            north: [],
	        south: []
    	};
    }

    this._screens.sort(function(s1, s2) {
        return s1.x - s2.x;
    });
}

/**
 * Returns the index of the screen, a specific window belongs to.
 */
function getCurrentScreenIndex(win) {
	let pos = win.get_outer_rect();
	pos.x = pos.x < 0 ? 0 : pos.x;

	let sl = this._screens.length;
  
	for (let i=0; i<sl; i++) {
		if (i == sl-1) {
			return i;
		}
		if (this._screens[i].x <= pos.x && this._screens[(i+1)].x > pos.x) {
			return i;
		}
    }
    
	return this._primary;
}

/**
 * Returns the index of the screen, which has been defined to contain the top
 * panel of the GNOME shell in the preferences dialog.
 */
function getPanelScreen() {
	switch (settings.get_enum(Prefs.PANEL_SCREEN_KEY)) {
	    case Prefs.PanelScreen.Primary_Screen: return this._primary;
	    case Prefs.PanelScreen.Screen_1: return 0;
	    case Prefs.PanelScreen.Screen_2: return 1;
	    case Prefs.PanelScreen.Screen_3: return 2;
	    case Prefs.PanelScreen.Screen_4: return 3;
        default: return this._primary;
    }
}

/**
 * Tells the window manager to hide the titlebar on maximised windows. TODO:
 * GNOME 3.2?
 * 
 * Note - no checking of blacklists etc is done in the function. You should do
 * it prior to calling the function (same with {@link decorate} and
 * {@link undecorate}).
 * 
 * Does this by setting the _GTK_HIDE_TITLEBAR_WHEN_MAXIMIZED hint - means I can
 * do it once and forget about it, rather than tracking maximize/unmaximize
 * events.
 * 
 * **Caveat**: doesn't work with Ubuntu's Ambiance and Radiance window themes -
 * my guess is they don't respect or implement this property.
 * 
 * @param {Meta.Window}
 *            win - window to set the HIDE_TITLEBAR_WHEN_MAXIMIZED property of.
 * @param {boolean}
 *            hide - whether to hide the titlebar or not.
 * @param {boolean}
 *            [stopAdding] - if `win` does not have an actor and we couldn't
 *            find the window's XID, we try one more time to detect the XID,
 *            unless this is `true`. Internal use.
 */
function setHideTitlebar(win, hide, stopAdding) {
	let screenIndex = getCurrentScreenIndex(win);
    let onlyPanelScreenWindows = settings.get_boolean(Prefs.PANEL_SCREEN_ONLY_KEY);
    let panelScreen = getPanelScreen();
	
	// When a window should be hidden, check if it really has to be hidden
	// depending on the screen settings.
	if (hide && onlyPanelScreenWindows && screenIndex != panelScreen) {
		hide = false;
	}
	
	LOG('setHideTitlebar: ' + win.get_title() + ': ' + hide
			+ (stopAdding ? ' (2)' : ''));

	let	id = guessWindowXID(win);
	/*
	 * Newly-created windows are added to the workspace before the compositor
	 * knows about them: get_compositor_private() is null. Additionally things
	 * like .get_maximized() aren't properly done yet. (see workspace.js
	 * _doAddWindow)
	 */
	if (!id && !win.get_compositor_private() && !stopAdding) {
		Mainloop.idle_add(function() {
			setHideTitlebar(null, win, true); // only try once more.
			return false; // define as one-time event
		});
		return;
	}

	/*
	 * Undecorate with xprop. Use _GTK_HIDE_TITLEBAR_WHEN_MAXIMIZED. See (eg)
	 * mutter/src/window-props.c
	 */
	let	cmd = [ 'xprop', '-id', id, '-f', '_GTK_HIDE_TITLEBAR_WHEN_MAXIMIZED',
			'32c', '-set', '_GTK_HIDE_TITLEBAR_WHEN_MAXIMIZED',
			(hide ? '0x1' : '0x0') ];

	// fallback: if couldn't get id for some reason, use the window's name
	if (!id) {
		cmd[1] = '-name';
		cmd[2] = win.get_title();
	}
	LOG(cmd.join(' '));
	Util.spawn(cmd);
}

/**
 * Returns whether we should affect `win`'s decorationa t all.
 * 
 * If the window was originally undecorated we do not do anything with it
 * (decorate or undecorate),
 * 
 * Also if it's in the blacklist, or if it's NOT in the whitelist, we don't do
 * anything with it.
 * 
 * @returns {boolean} whether the window is originally decorated and not in the
 *          blacklist (or in the whitelist).
 */
function shouldAffect(win) {
	if (!win._decoratedOriginal) {
		return false;
	}
	let	app = Shell.WindowTracker.get_default().get_window_app(win),
			appid = (app ? app.get_id() : -1), inList = APP_LIST.length > 0
					&& APP_LIST.indexOf(appid) >= 0;
	return !((IS_BLACKLIST && inList) || (!IS_BLACKLIST && !inList));
}

/**
 * Checks if `win` should be undecorated, based *purely* off its maximised state
 * (doesn't incorporate blacklist).
 * 
 * If it's fully-maximized or half-maximised and undecorateHalfMaximised is
 * true, this returns true.
 * 
 * Use with `shouldAffect` to get a full check..
 */
function shouldBeUndecorated(win) {
	let	max = win.get_maximized();
	return ((max === Meta.MaximizeFlags.BOTH) || (settings
			.get_boolean(Prefs.UNDECORATE_HALF_MAXIMIZED_KEY) && max));
}

/**
 * Checks if `win` is fully maximised, or half-maximised +
 * undecorateHalfMaximised. If so, undecorates the window.
 */
function possiblyUndecorate(win) {
	if (shouldBeUndecorated(win)) {
		if (!win.get_compositor_private()) {
			Mainloop.idle_add(function() {
				undecorate(win);
				return false; // define as one-time event
			});
		} else {
			undecorate(win);
		}
	}
}

/**
 * Checks if `win` is fully maximised, or half-maximised +
 * undecorateHalfMaximised. If *NOT*, redecorates the window.
 */
function possiblyRedecorate(win) {
	if (!shouldBeUndecorated(win)) {
		if (!win.get_compositor_private()) {
			Mainloop.idle_add(function() {
				decorate(win);
				return false; // define as one-time event
			});
		} else {
			decorate(win);
		}
	}
}

/** ** Callbacks *** */
/**
 * Called when a window is maximized, including half-maximization.
 * 
 * If the window is not in the blacklist (or is in the whitelist), we undecorate
 * it.
 * 
 * @param {Meta.WindowActor}
 *            actor - the window actor for the maximized window. It is expected
 *            to be maximized (in at least one direction) already - we will not
 *            check before undecorating.
 */
function onMaximise(shellwm, actor) {
	if (!actor) {
		return;
	}
	let	win = actor.get_meta_window();
	if (!shouldAffect(win)) {
		return;
	}
	// note: window is maximized by this point.
	let	max = win.get_maximized();
	LOG('onMaximise: ' + win.get_title() + ' [' + win.get_wm_class() + ']');
	// if this is a partial maximization, and we do not wish to undecorate
	// half-maximized windows, make sure the window is decorated.
	if (max !== Meta.MaximizeFlags.BOTH
			&& !settings.get_boolean(Prefs.UNDECORATE_HALF_MAXIMIZED_KEY)) {
		decorate(win);
		return;
	}
	undecorate(win);
}

/**
 * Called when a window is unmaximized.
 * 
 * If the window is not in the blacklist (or is in the whitelist), we decorate
 * it.
 * 
 * @param {Meta.WindowActor}
 *            actor - the window actor for the unmaximized window. It is
 *            expected to be unmaximized - we will not check before decorating.
 */
function onUnmaximise(shellwm, actor) {
	if (!actor) {
		return;
	}
	let	win = actor.meta_window;
	if (!shouldAffect(win)) {
		return;
	}
	LOG('onUnmaximise: ' + win.get_title());
	// if the user is unmaximizing by dragging, we wait to decorate until they
	// have dropped the window, so that we don't force the user to drop
	// the window prematurely with the redecorate (which stops the grab).
	//
	// This is only necessary if USE_SET_HIDE_TITLEBAR is `false` (otherwise
	// this is not an issue).
	if (!USE_SET_HIDE_TITLEBAR
			&& global.display.get_grab_op() === Meta.GrabOp.MOVING) {
		if (grabID) {
			// shouldn't happen, but oh well.
			global.display.disconnect(grabID);
			grabID = 0;
		}
		grabID = global.display.connect('grab-op-end', function() {
			possiblyRedecorate(win);
			global.display.disconnect(grabID);
			grabID = 0;
		});
	} else {
		decorate(win);
	}
}

/**
 * Callback for a window's 'notify::maximized-horizontally' and
 * 'notify::maximized-vertically' signals.
 * 
 * If the window is half-maximised we force it to show its titlebar. Otherwise
 * we set it to hide if it is maximized.
 * 
 * Only used if using the SET_HIDE_TITLEBAR method AND we wish half-maximized
 * windows to be *decorated* (the GTK_HIDE_TITLEBAR_WHEN_MAXIMIZED atom will
 * hide the titlebar of half-maximized windows too).
 * 
 * @param {Meta.Window}
 *            win - the window whose maximized-horizontally or
 *            maximized-vertically properties has changed.
 * 
 * @see onWindowAdded
 */
function onWindowChangesMaximiseState(win) {
	if ((win.maximized_horizontally && !win.maximized_vertically)
			|| (!win.maximized_horizontally && win.maximized_vertically)) {
		setHideTitlebar(win, false);
		decorate(win);
	} else {
		setHideTitlebar(win, true);
	}
}

/**
 * Callback when a window is added in any of the workspaces. This includes a
 * window switching to another workspace.
 * 
 * If it is a window we already know about, we do nothing.
 * 
 * Otherwise, we: * record the window as on we know about. * store whether the
 * window was initially decorated (e.g. Chrome windows aren't usually). * if
 * using the SET_HIDE_TITLEBAR method, we: + set the
 * GTK_HIDE_TITLEBAR_WHEN_MAXIMIZED atom on the window. + if we wish to keep
 * half-maximised windows decorated, we connect up some signals to ensure that
 * half-maximised windows remain decorated (the GTK_HIDE_TITLEBAR_WHEN_MAXIMIZED
 * atom will automatically undecorated half-maximised windows). See
 * {@link onWindowChangesMaximiseState}. * otherwise (not using
 * SET_HIDE_TITLEBAR): + if the window is maximized, we undecorate it (see
 * {@link undecorate}); + if the window is half-maximized and we wish to
 * undecorate half-maximised windows, we also undecorate it.
 * 
 * @param {Meta.Window}
 *            win - the window that was added.
 * 
 * @see undecorate
 */
function onWindowAdded(ws, win) {
	// if the window is simply switching workspaces, it will trigger a
	// window-added signal. We don't want to reprocess it then because we
	// already
	// have.
	if (win._decoratedOriginal !== undefined) {
		return;
	}

	/*
	 * Newly-created windows are added to the workspace before the compositor
	 * knows about them: get_compositor_private() is null. Additionally things
	 * like .get_maximized() aren't properly done yet. (see workspace.js
	 * _doAddWindow)
	 */
	win._decoratedOriginal = win.decorated !== false || false;
	LOG('onWindowAdded: ' + win.get_title() + ' initially decorated? '
			+ win._decoratedOriginal);

	if (!shouldAffect(win)) {
		return;
	}

	// with set_hide_titlebar, set the window hint when the window is added and
	// there is no further need to listen to maximize/unmaximize on the window.
	if (USE_SET_HIDE_TITLEBAR) {
		setHideTitlebar(win, true);
		// set_hide_titlebar undecorates half maximized, so if we wish not to we
		// will have to manually redo it ourselves
		if (!settings.get_boolean(Prefs.UNDECORATE_HALF_MAXIMIZED_KEY)) {
			win._maxHStateId = win.connect('notify::maximized-horizontally',
					onWindowChangesMaximiseState);
			win._maxVStateId = win.connect('notify::maximized-vertically',
					onWindowChangesMaximiseState);
			if (win.get_maximized()) {
				onWindowChangesMaximiseState(win);
			}
		}
	} else {
		// if it is added initially maximized, we undecorate it.
		possiblyUndecorate(win);
	}
}

/**
 * Callback whenever the number of workspaces changes.
 * 
 * We ensure that we are listening to the 'window-added' signal on each f the
 * workspaces.
 * 
 * @see onWindowAdded
 */
function onChangeNWorkspaces() {
	let	ws, i = workspaces.length;
	while (i--) {
		workspaces[i].disconnect(workspaces[i]._windowAddedId);
	}

	workspaces = [];
	i = global.screen.n_workspaces;
	while (i--) {
		ws = global.screen.get_workspace_by_index(i);
		workspaces.push(ws);
		// we need to add a Mainloop.idle_add, or else in onWindowAdded the
		// window's maximized state is not correct yet.
		// ws._windowAddedId = ws.connect('window-added', onWindowAdded);
		ws._windowAddedId = ws.connect('window-added', function(ws, win) {
			Mainloop.idle_add(function() {
				onWindowAdded(ws, win);
				return false;
			})
		});
	}
}

/** Start listening to events and undecorate already-existing windows. */
function startUndecorating() {
	// cache some variables for convenience
	IS_BLACKLIST = settings.get_boolean(Prefs.IS_BLACKLIST_KEY);
	APP_LIST = settings.get_strv(Prefs.BLACKLIST_KEY);
	USE_SET_HIDE_TITLEBAR = settings
			.get_boolean(Prefs.USE_SET_HIDE_TITLEBAR_KEY);
	if (USE_SET_HIDE_TITLEBAR
			&& Meta.prefs_get_theme().match(/^(?:Ambiance|Radiance)$/)) {
		USE_SET_HIDE_TITLEBAR = false;
	}

	/* Connect events */
	changeWorkspaceID = global.screen.connect('notify::n-workspaces',
			onChangeNWorkspaces);
	// if we are not using the set_hide_titlebar hint, we must listen to
	// maximize and unmaximize events.
	if (!USE_SET_HIDE_TITLEBAR) {
		maxID = global.window_manager.connect('maximize', onMaximise);
		minID = global.window_manager.connect('unmaximize', onUnmaximise);
		/*
		 * this is needed to prevent Metacity from interpreting an attempted
		 * drag of an undecorated window as a fullscreen request. Otherwise
		 * thunderbird (in particular) has no way to get out of fullscreen,
		 * resulting in the user being stuck there.
		 * 
		 * Once we can properly set the window's hide_titlebar_when_maximized
		 * property this will no loner be necessary.
		 */
		oldFullscreenPref = Meta.prefs_get_force_fullscreen();
		Meta.prefs_set_force_fullscreen(false);
	}

	/*
	 * Go through already-maximised windows & undecorate. This needs a delay as
	 * the window list is not yet loaded when the extension is loaded. Also,
	 * connect up the 'window-added' event. Note that we do not connect this
	 * before the onMaximise loop because when one restarts the gnome-shell,
	 * window-added gets fired for every currently-existing window, and then
	 * these windows will have onMaximise called twice on them.
	 */
	onetime = Mainloop.idle_add(function() {
		let	winList = global.get_window_actors().map(function(w) {
			return w.meta_window;
		}), i = winList.length;
		while (i--) {
			let	win = winList[i];
			if (win.window_type === Meta.WindowType.DESKTOP) {
				continue;
			}
			onWindowAdded(null, win);
		}
		onChangeNWorkspaces();
		return false; // define as one-time event
	});
}

/**
 * Stop listening to events, restore all windows back to their original
 * decoration state.
 */
function stopUndecorating() {
	if (maxID)
		global.window_manager.disconnect(maxID);
	if (minID)
		global.window_manager.disconnect(minID);
	if (changeWorkspaceID)
		global.window_manager.disconnect(changeWorkspaceID);
	if (grabID)
		global.display.disconnect(grabID);
	maxID = 0;
	minID = 0;
	changeWorkspaceID = 0;
	grabID = 0;

	/* disconnect window-added from workspaces */
	let	i = workspaces.length;
	while (i--) {
		workspaces[i].disconnect(workspaces[i]._windowAddedId);
		delete workspaces[i]._windowAddedId;
	}
	workspaces = [];

	/* redecorate undecorated windows we screwed with */
	if (onetime) {
		Mainloop.source_remove(onetime);
		onetime = 0;
	}
	let	winList = global.get_window_actors().map(function(w) {
		return w.meta_window;
	}), i = winList.length;
	while (i--) {
		let	win = winList[i];
		if (win.window_type === Meta.WindowType.DESKTOP) {
			continue;
		}
		LOG('stopUndecorating: ' + win.title);
		// if it wasn't decorated originally, we haven't done anything to it so
		// don't need to undo anything.
		if (win._decoratedOriginal) {
			if (USE_SET_HIDE_TITLEBAR) {
				setHideTitlebar(win, false);
				if (win._maxHStateId) {
					win.disconnect(win._maxHStateId);
					delete win._maxHStateId;
				}
				if (win._maxVStateId) {
					win.disconnect(win._maxVStateId);
					delete win._maxVStateId;
				}
			}
			decorate(win);
		}
		delete win._decoratedOriginal;
	}

	if (oldFullscreenPref !== null) {
		/* restore old meta force fullscreen pref */
		Meta.prefs_set_force_fullscreen(oldFullscreenPref);
		oldFullscreenPref = null;
	}
}

function init() {
	settings = Convenience.getSettings();
	loadScreenData();
}

function enable() {
	startUndecorating();

	/* Monitor settings changes */
	settingsChangedID = settings.connect('changed', function() {
		// redecorate every window and undecorate again according to the
		// new settings.
		stopUndecorating();

		startUndecorating();
	});
}

function disable() {
	stopUndecorating();

	/* disconnect settings changes */
	if (settingsChangedID) {
		settings.disconnect(settingsChangedID);
	}
}
